INIT_KEY = "__init"
FN_KEY = "__fn"
TARGET_KEY = "__target"
VALIDATE_KEY = "__validate"
PARTIAL_KEY = "__partial"
ARGS_KEY = "__args"
KEY_KEY = "__key"
PLACEHOLDER_KEY = "__placeholder"
META_KEY = "__meta"
ANNOTATE_KEY = "__annotate"
TAG_KEY = "__tag"

EXCLUDE_KEYS = {
    INIT_KEY,
    TARGET_KEY,
    FN_KEY,
    VALIDATE_KEY,
    PARTIAL_KEY,
    ARGS_KEY,
    KEY_KEY,
    META_KEY,
}
